/*
 * Copyright 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is part of libstrl.
 *
 * libstrl is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * libstrl is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with libstrl.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "suite_getdelim.h"

#include <check.h>
#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <strl.h>

START_TEST (getdelim_errno)
{
  ssize_t ret;
  char *lineptr;
  size_t n;
  FILE *in;

  size_t counter;

  /*
   * .stream is set up lazily in the lower for loop
   *
   * .stream = NULL isn't handled by GNU getline(), so when we're
   * running this test against the system getline() we don't ever pass
   * stream = NULL to it. I still think this is a valid thing to test
   * for, based on the manpage's description of when EINVAL would be
   * set. --binki
   */
  struct getdelim_args
  {
    char **lineptr;
    size_t *n;
    FILE *stream;
  } getdelim_args[] = {
#ifdef _LIBSTRL_DECLARE_GETDELIM
    {.lineptr = NULL    , .n = NULL, .stream = NULL},
#endif /* _LIBSTRL_DECLARE_GETDELIM */
    {.lineptr = NULL    , .n = &n  , .stream = (void *)0x01},
    {.lineptr = &lineptr, .n = NULL, .stream = (void *)0x01},
#ifdef _LIBSTRL_DECLARE_GETDELIM
    {.lineptr = &lineptr, .n = &n  , .stream = NULL},
#endif /* _LIBSTRL_DECLARE_GETDELIM */
  };

  in = fopen(GETLINE_ENDNEWLINE_PATH, "rb");
  fail_if(!in, "Unable to open ``" GETLINE_ENDNEWLINE_PATH "'' for the purpose of testing getdelim().");

  for (counter = 0; counter < sizeof(getdelim_args) / sizeof(struct getdelim_args); counter ++)
    {
      /* hacky way to lazily set up getdelim_args ;-) */
      if (getdelim_args[counter].stream)
	getdelim_args[counter].stream = in;

      errno = 0;
      ret = getdelim(getdelim_args[counter].lineptr, getdelim_args[counter].n, ',', getdelim_args[counter].stream);
      fail_unless(ret == -1, "getdelim() expected to return -1, returned %zd instead. Argument index %zu.",
		  ret, counter);
      fail_unless(errno == EINVAL);
    }

  fclose(in);
}
END_TEST

START_TEST (getdelim_eof)
{
  ssize_t ret;
  char *lineptr;
  size_t n;
  FILE *in;

  short has_hit_eof;

  in = fopen(GETLINE_ENDABRUPT_PATH, "rb");
  fail_if(!in, "Unable to open ``" GETLINE_ENDABRUPT_PATH "'' for the purpose of testing getdelim().");

  lineptr = NULL;
  has_hit_eof = 0;
  while ((ret = getdelim(&lineptr, &n, '\n', in)) != -1)
    {
      fail_if (has_hit_eof && feof(in),
	       "getdelim() returned %zd instead of -1 when the end-of-file condition was reached (for file _not_ terminated with newline).", ret);

      /*
       * The first time we hit the EOF, we read valid data in at the
       * same time. This is because this is the file which isn't
       * terminated with a delimiter. Therefor, this loop must see
       * feof(in) be true for two loops for this to be a bug.
       */
      if (feof(in))
	has_hit_eof = 1;
    }
  fail_unless(feof(in),
	      "getdelim() returned %zd even though the end of the file was not yet reached (for file _not_ terminated with newline).", ret);

  fclose(in);

  /*
   * Try with when the file does end with a delimiter: we should
   * _never_ have feof(in) be true inside of _this_ loop.
   */
  in = fopen(GETLINE_ENDNEWLINE_PATH, "rb");
  fail_if(!in, "Unable to open ``" GETLINE_ENDNEWLINE_PATH "'' for the purpose of testing getdelim().");

  while ((ret = getdelim(&lineptr, &n, '\n', in)) != -1)
    fail_if(feof(in),
	    "getdelim() returned %zd instead of -1 when the end-of-file condition was reached (for file terminated with newline).", ret);

  fail_unless(feof(in),
	      "getdelim() returned %zd even though the end of the file was not yet reached (for file terminated with newline).", ret);

  fclose(in);

  free(lineptr);
}
END_TEST

Suite *getdelim_suite()
{
  Suite *s;
  TCase *tc;

  s = suite_create("getdelim");

  tc = tcase_create("errno");
  tcase_add_test(tc, getdelim_errno);
  suite_add_tcase(s, tc);

  tc = tcase_create("eof");
  tcase_add_test(tc, getdelim_eof);
  suite_add_tcase(s, tc);

  return s;
}
