/*
 * Copyright 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is part of libstrl.
 *
 * libstrl is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * libstrl is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with libstrl.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "suite_getline.h"

#include <check.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strl.h>

START_TEST (getline_normal)
{
  char *lineptr;
  size_t n;
  ssize_t ret;

  FILE *in;

  /*
   * Coordinate this value with the lengths of the shipped files to
   * ensure that we don't buffer-overflow in the tests.
   */
  char inbuf[1024];
  size_t inbuf_len;
  size_t inbuf_pos;

  in = fopen(GETLINE_ENDNEWLINE_PATH, "rb");
  fail_if(in == NULL, "Unable to open file required for test: ``" GETLINE_ENDNEWLINE_PATH "''");

  inbuf_len = fread(&inbuf, 1, sizeof(inbuf), in);
  fail_unless(inbuf_len,
	      "Unable to read data from ``" GETLINE_ENDNEWLINE_PATH "''");

  fclose(in);
  in = fopen(GETLINE_ENDNEWLINE_PATH, "rb");
  fail_if(in == NULL, "Unable to open file required for test: ``" GETLINE_ENDNEWLINE_PATH "''");

  inbuf_pos = 0;
  inbuf[inbuf_len] = '\0';

  lineptr = NULL;
  while(inbuf_pos < inbuf_len)
    {
      ret = getline(&lineptr, &n, in);
      fail_if(ret == -1, "getline() returned -1 when unexpected, still had the following data to read: %s",
	      &inbuf[inbuf_pos]);
      fail_unless(strlen(lineptr) == ret, "getline()'s return value did not match the length of the string it stored."
		  " strlen(lineptr) = %zu, ret = %zu.", strlen(lineptr), ret);
      fail_unless(!strncmp(lineptr, &inbuf[inbuf_pos], ret),
		  "getline()'s returned string did not match. First chars: `%c', `%c'.",
		  inbuf[inbuf_pos], *lineptr);
      fail_if(strchr(lineptr, '\n') < lineptr + ret - 1, "getline() skipped a delimiter.");
      inbuf_pos += ret;
    }
  free(lineptr);

  fclose(in);
}
END_TEST

START_TEST (getline_early_eof)
{
  char *lineptr;
  size_t n;
  ssize_t ret;

  FILE *in;

  /*
   * Coordinate this value with the lengths of the shipped files to
   * ensure that we don't buffer-overflow in the tests.
   */
  char inbuf[1024];
  size_t inbuf_len;
  size_t inbuf_pos;

  char mydelim;

  in = fopen(GETLINE_ENDABRUPT_PATH, "rb");
  fail_if(in == NULL, "Unable to open file required for test: ``" GETLINE_ENDABRUPT_PATH "''");

  inbuf_len = fread(&inbuf, 1, sizeof(inbuf), in);
  fail_unless(inbuf_len,
	      "Unable to read data from ``" GETLINE_ENDABRUPT_PATH "''");

  fclose(in);
  in = fopen(GETLINE_ENDABRUPT_PATH, "rb");
  fail_if(in == NULL, "Unable to open file required for test: ``" GETLINE_ENDABRUPT_PATH "''");

  inbuf_pos = 0;
  inbuf[inbuf_len] = '\0';

  lineptr = NULL;
  while(inbuf_pos < inbuf_len)
    {
      ret = getline(&lineptr, &n, in);
      fail_if(ret == -1, "getline() returned -1 when unexpected, still had the following data to read: %s",
	      &inbuf[inbuf_pos]);
      fail_unless(strlen(lineptr) == ret, "getline()'s return value did not match the length of the string it stored."
		  " strlen(lineptr) = %zu, ret = %zu.", strlen(lineptr), ret);
      fail_unless(!strncmp(lineptr, &inbuf[inbuf_pos], ret),
		  "getline()'s returned string did not match. First chars: `%c', `%c'.",
		  inbuf[inbuf_pos], *lineptr);

      mydelim = '\n';
      if (feof(in))
	mydelim = '\0';
      fail_if(strchr(lineptr, mydelim) < lineptr + ret - 1, "getline() skipped a delimiter.");
      inbuf_pos += ret;
    }
  free(lineptr);

  fclose(in);
}
END_TEST

START_TEST (getline_long_line)
{
  char *lineptr;
  size_t n;
  ssize_t ret;

  FILE *in;

  lineptr = NULL;

  in = fopen(GETLINE_LONGLINE_PATH, "rb");
  fail_if(in == NULL, "Unable to open file required for test: ``" GETLINE_LONGLINE_PATH "''");

  ret = getline(&lineptr, &n, in);
  fail_if(ret == -1, "getline() returned -1 when unexpected");
  fail_unless(strlen(lineptr) == ret, "getline()'s return value did not match the length of the string it stored."
	      " strlen(lineptr) = %zu, ret = %zd, data: ``%s''.", strlen(lineptr), ret, lineptr);
  fail_unless(strlen(lineptr) > 80, "getline() did not return all of the data in ``" GETLINE_LONGLINE_PATH "''. Got data: ``%s''.",
	      lineptr);

  free(lineptr);
  fclose(in);
}
END_TEST

Suite *getline_suite()
{
  Suite *s;
  TCase *tc;

  s = suite_create("getline");

  tc = tcase_create("normal");
  tcase_add_test(tc, getline_normal);
  suite_add_tcase(s, tc);

  tc = tcase_create("early_eof");
  tcase_add_test(tc, getline_early_eof);
  suite_add_tcase(s, tc);

  tc = tcase_create("long_line");
  tcase_add_test(tc, getline_long_line);
  suite_add_tcase(s, tc);

  return s;
}
