/*
 * Copyright 2010 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is part of libstrl.
 *
 * libstrl is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * libstrl is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with libstrl.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "suite_strlcat.h"

#include <check.h>
#include <stdlib.h>
#include <string.h>
#include <strl.h>

START_TEST (strlcat_empty)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(len);
  fail_unless(dst != 0, "Unable to malloc().");
  dst[0] = '\0';

  ret = strlcat(dst, src, len);
  fail_unless(ret == strlen(src),
	      "ret == %d, strlen(src) == %d: should be equal.",
	      ret, strlen(src));
  /* no truncation should have occured */
  fail_unless(ret == strlen(dst),
	      "Expecting return value of %d, got %d", strlen(dst), ret);
  fail_unless(dst[ret] == '\0',
	      "String not properly terminated with a null char");
  fail_unless(!strcmp(dst, src),
	      "Expecting concatenation to be `%s', got `%s'",
	      src, dst);

  free(dst);
}
END_TEST

START_TEST (strlcat_append)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 20;

  dst = malloc(len);
  fail_unless(dst != 0, "Unable to malloc().");

  /* seed: */
  memcpy(dst, "food: ", strlen("food: ") + 1);
  fail_unless(!strcmp(dst, "food: "),
	      "Unable to seed dst with ``food: ''");

  ret = strlcat(dst, src, len);
  fail_unless(ret == strlen("food: ") + strlen(src),
	      "Expecting string length of %d, got %d",
	      strlen("food: ") + strlen(src));
  /* no truncation should have occured */
  fail_unless(ret == strlen(dst),
	      "ret == %d, strlen(dst) == %d");
  fail_unless(dst[ret] == '\0',
	      "String not properly terminated with a null char");
  fail_unless(!strcmp(dst, "food: pizza"),
	      "Expecting ``food: pizza'', got something else");
}
END_TEST

START_TEST (strlcat_fill)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(len);
  fail_unless(dst != 0, "Unable to malloc()");

  /* seed, leaving enough room for pizza and the terminator */
  memcpy(dst, "abcdefghijklmnopqrstuvwxyz", len - strlen(src) - 1);
  dst[len - 1] = '\0';
  dst[len - strlen(src) - 1] = '\0';
  fail_unless(!strcmp(dst, "abcdefghijklmn"),
	      "Unable to properly seed dst: %s", dst);

  ret = strlcat(dst, src, len);
  fail_unless(ret == 19,
	      "ret == %d, expected 19", ret);
  /* no truncation should have occured */
  fail_unless(ret == strlen(dst),
	      "ret == %d, strlen(dst) == %d");
  fail_unless(ret == len - 1,
	      "Expected return value to be %d, got %d",
	      len, ret);
  fail_unless(!strcmp(dst, "abcdefghijklmnpizza"),
	      "Got something other than ``abcdefghijklmnpizza''.");
}
END_TEST

START_TEST (strlcat_clip)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 10;

  dst = malloc(len);
  fail_unless(dst != 0,
	      "Unable to malloc()");

  /* seed */
  memcpy(dst, "I love ", strlen("I love ") + 1);
  fail_unless(!strncmp(dst, "I love ", strlen("I love ") + 1),
	      "Unable to seed");

  ret = strlcat(dst, src, len);
  fail_unless(ret == strlen("I love ") + strlen(src),
	      "Expected ret == strlen(L\"I love \") + strlen(src) which is %d, instead ret = %d",
	      strlen("I love ") + strlen(src), ret);
  fail_unless(dst[len - 1] == '\0',
	      "Expected terminating NULL char, instead got `%c'",
	      (char)dst[len - 1]);
  fail_unless(!strcmp(dst, "I love pi"),
	      "Expected dst to be ``I love pi'', got something else");
}
END_TEST

START_TEST (strlcat_overfull)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(len);
  fail_unless(dst != 0, "Unable to malloc()");

  /* seed without a NULL byte */
  memcpy(dst, "12345678901234567890", len);
  fail_unless(dst[len - 1] == '0',
	       "Failed to properly seed dst; dst[%d] = %c.",
	       len - 1, dst[len - 1]);

  /*
   * In this case, we supposedly didn't properly terminate dst with a
   * NULL byte. Thus, strlcat() is supposed to leave this last byte
   * untouched.
   */
  ret = strlcat(dst, src, len);
  fail_unless(ret == len + strlen(src),
	      "Expected ret == len + strlen(src), but ret = %d and len + strlen(src) = %d",
	      ret, len + strlen(src));
  fail_unless(!strncmp(dst, "12345678901234567890", len),
	      "Expected dst not to be modified, it was");
  fail_unless(dst[len - 1] != '\0',
	      "Expected last byte of dst not to be '\0', but it was '\0'");
}
END_TEST

START_TEST (strlcat_zero)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 0;

  dst = malloc(len);
  /*
   * It is valid for dst to be either NULL or some valid pointer at
   * this point.
   */

  ret = strlcat(dst, src, len);
  fail_unless(ret == strlen(src),
	      "Expected ret to be %d, was instead %d.",
	      strlen(src), ret);

  free(dst);

  len ++;
  dst = malloc(len);
  fail_unless(dst != 0,
	      "Unable to malloc() %d bytes", len);

  dst[0] = '\0';

  ret = strlcat(dst, src, len);
  fail_unless(ret == strlen(src),
	      "Expected ret to be %d, was instead %d.",
	      strlen(src), ret);
  fail_unless(dst[0] == '\0',
	      "Destination string was not properly null-terminated, last char is `%c'.",
	      dst[0]);
  fail_unless(strlen(dst) == len - 1,
	      "Destination string's length is %d, expected length of %d",
	      strlen(dst), len - 1);

  free(dst);
}
END_TEST

Suite *strlcat_suite()
{
  Suite *s;
  TCase *tc;

  s = suite_create("strlcat");

  tc = tcase_create("empty");
  tcase_add_test(tc, strlcat_empty);
  suite_add_tcase(s, tc);

  tc = tcase_create("append");
  tcase_add_test(tc, strlcat_append);
  suite_add_tcase(s, tc);

  tc = tcase_create("fill");
  tcase_add_test(tc, strlcat_fill);
  suite_add_tcase(s, tc);

  tc = tcase_create("clip");
  tcase_add_test(tc, strlcat_clip);
  suite_add_tcase(s, tc);

  tc = tcase_create("overfull");
  tcase_add_test(tc, strlcat_overfull);
  suite_add_tcase(s, tc);

  tc = tcase_create("zero");
  tcase_add_test(tc, strlcat_zero);
  suite_add_tcase(s, tc);

  return s;
}
