/*
 * Copyright 2010 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is part of libstrl.
 *
 * libstrl is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * libstrl is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with libstrl.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "suite_strlcpy.h"

#include <check.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <strl.h>

START_TEST (strlcpy_empty)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(len);
  fail_unless(dst != 0, "Unable to malloc()");
  dst[0] = '\0';

  ret = strlcpy(dst, src, len);
  fail_unless(ret == strlen(src),
	      "Expected return  value of %d, was instead %d", strlen(src), ret);
  fail_unless(strlen(src) == strlen(dst),
	      "Source and destination lengths differ: source: %d, dest: %d",
	      strlen(src), strlen(dst));
  fail_unless(dst[ret] == '\0',
	      "Destination not properly null-terminated");
  fail_unless(!strcmp(dst, src),
	      "Expected `%s', got something else.", src);

  free(dst);
}
END_TEST

START_TEST (strlcpy_full)
{
  char *dst;
  const char *src = "abcdefghiJklmnopqr";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(len);
  fail_unless(dst != 0, "Unable to malloc()");

  ret = strlcpy(dst, src, len);
  fail_unless(ret == strlen(src),
	      "Expected return value of %d, was instead %d", strlen(src), ret);
  fail_unless(strlen(src) == strlen(dst),
	      "Source and destination lengths differ: source %d, dest: %d",
	      strlen(src), strlen(dst));
  fail_unless(dst[ret] == '\0',
	      "Destination not properly null-terminated");
  fail_unless(!strcmp(dst, src),
	      "Expected long version of `abcdefghiJklmnopqr', got something else.");

  free(dst);
}
END_TEST

START_TEST(strlcpy_zero)
{
  char *dst;
  const char *src = "pizza";
  size_t len;
  size_t ret;

  len = 0;

  dst = malloc(len);
  /*
   * It is valid for dst to be either NULL or some valid pointer at
   * this point.
   */

  ret = strlcpy(dst, src, len);
  fail_unless(ret == strlen(src),
	      "Expected ret to be %d, was instead %d.",
	      strlen(src), ret);

  free(dst);

  len ++;
  dst = malloc(len);
  fail_unless(dst != 0,
	      "Unable to malloc() %d bytes", len);

  ret = strlcpy(dst, src, len);
  fail_unless(ret == strlen(src),
	      "Expected ret to be %d, was instead %d.",
	      strlen(src), ret);
  fail_unless(dst[0] == '\0',
	      "Destination string was not properly null-terminated, last char is `%c'.",
	      dst[0]);
  fail_unless(strlen(dst) == len - 1,
	      "Expected the string length of dst to be %d, was instead %d",
	      len - 1, strlen(dst));

  free(dst);
}
END_TEST

Suite *strlcpy_suite()
{
  Suite *s;
  TCase *tc;

  s = suite_create("strlcpy");

  tc = tcase_create("empty");
  tcase_add_test(tc, strlcpy_empty);
  suite_add_tcase(s, tc);

  tc = tcase_create("full");
  tcase_add_test(tc, strlcpy_full);
  suite_add_tcase(s, tc);

  tc = tcase_create("zero");
  tcase_add_test(tc, strlcpy_zero);
  suite_add_tcase(s, tc);

  return s;
}
