/*
 * Copyright 2010 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is part of libstrl.
 *
 * libstrl is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * libstrl is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with libstrl.  If not, see
 * <http://www.gnu.org/licenses/>.
 */

#include "suite_wcslcat.h"

#include <check.h>
#include <stdlib.h>
#include <string.h>
#include <strl.h>
#include <wchar.h>

START_TEST (wcslcat_empty)
{
  wchar_t *dst;
  const wchar_t *src = L"pizza";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(sizeof(wchar_t) * len);
  fail_unless(dst != 0, "Unable to malloc().");
  dst[0] = L'\0';

  ret = wcslcat(dst, src, len);
  fail_unless(ret == wcslen(src),
	      "Expecting return value of %d, got %d", wcslen(src), ret);
  /* no truncation should have occured */
  fail_unless(ret == wcslen(dst),
	      "ret == %d, wclen(dst) == %d",
	      ret, wcslen(dst));
  fail_unless(dst[ret] == L'\0',
	      "Wcsing not properly terminated with a null char");
  fail_unless(!wcscmp(dst, src),
	      "Expecting concatenation to be a certain wide string, got some other wide string instead",
	      src, dst);

  free(dst);
}
END_TEST

START_TEST (wcslcat_append)
{
  wchar_t *dst;
  const wchar_t *src = L"pizza";
  size_t len;
  size_t ret;

  len = 20;

  dst = malloc(sizeof(wchar_t) * len);
  fail_unless(dst != 0, "Unable to malloc().");

  /* seed: */
  memcpy(dst, L"food: ", sizeof(wchar_t) * (wcslen(L"food: ") + 1));
  fail_unless(!wcscmp(dst, L"food: "),
	      "Unable to seed dst with a wide version of ``food: ''");

  ret = wcslcat(dst, src, len);
  fail_unless(ret == wcslen(L"food: ") + wcslen(src),
	      "Expecting wcsing length of %d, got %d",
	      wcslen(L"food: ") + wcslen(src));
  /* no truncation should have occured */
  fail_unless(ret == wcslen(dst),
	      "ret == %d, wclen(dst) == %d",
	      ret, wcslen(dst));
  fail_unless(dst[ret] == L'\0',
	      "Wcsing not properly terminated with a null char");
  fail_unless(!wcscmp(dst, L"food: pizza"),
	      "Expecting a longer version of ``food: pizza'', got something near ``%c%c%c%c%c%c%c%c%c%c%c%c''",
	      dst[0], dst[1], dst[2], dst[3], dst[4], dst[5], dst[6],
	      dst[7], dst[8], dst[9]);
}
END_TEST

START_TEST (wcslcat_fill)
{
  wchar_t *dst;
  const wchar_t *src = L"pizza";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(sizeof(wchar_t) * len);
  fail_unless(dst != 0, "Unable to malloc()");

  /* seed, leaving enough room for pizza and the terminator */
  memcpy(dst, L"abcdefghijklmnopqrstuvwxyz", sizeof(wchar_t) * (len - wcslen(src) - 1));
  dst[len - 1] = L'\0';
  dst[len - wcslen(src) - 1] = L'\0';
  fail_unless(!wcscmp(dst, L"abcdefghijklmn"),
	      "Unable to properly seed dst: %s", dst);

  ret = wcslcat(dst, src, len);
  fail_unless(ret == 19,
	      "ret == %d, expected 19", ret);
  fail_unless(ret == len - 1,
	      "Expected return value to be %d, got %d",
	      len, ret);
  fail_unless(ret == wcslen(dst),
	      "ret == %d, wclen(dst) == %d",
	      ret, wcslen(dst));
  fail_unless(!wcscmp(dst, L"abcdefghijklmnpizza"),
	      "Got something other than the long version of ``abcdefghijklmnpizza''.");
}
END_TEST

START_TEST (wcslcat_clip)
{
  wchar_t *dst;
  const wchar_t *src = L"pizza";
  size_t len;
  size_t ret;

  len = 10;

  dst = malloc(len * sizeof(wchar_t));
  fail_unless(dst != 0,
	      "Unable to malloc()");

  /* seed */
  memcpy(dst, L"I love ", sizeof(wchar_t) * (wcslen(L"I love ") + 1));
  fail_unless(!wcsncmp(dst, L"I love ", wcslen(L"I love ") + 1),
	      "Unable to seed");

  ret = wcslcat(dst, src, len);
  fail_unless(ret == wcslen(L"I love ") + wcslen(src),
	      "Expected ret == wcslen(L\"I love \") + wcslen(src) which is %d, instead ret = %d",
	      wcslen(L"I love ") + wcslen(src), ret);
  fail_unless(dst[len - 1] == L'\0',
	      "Expected terminating NULL char, instead got a long char starting with `%c'",
	      (char)dst[len - 1]);
  fail_unless(!wcscmp(dst, L"I love pi"),
	      "Expected dst to have the long version of ``I love pi'', got something else");
}
END_TEST

START_TEST (wcslcat_overfull)
{
  wchar_t *dst;
  const wchar_t *src = L"pizza";
  size_t len;
  size_t ret;

  len = 20;
  dst = malloc(sizeof(wchar_t) * len);
  fail_unless(dst != 0, "Unable to malloc()");

  /* seed without a NULL byte */
  memcpy(dst, L"12345678901234567890", sizeof(wchar_t) * len);
  fail_unless(dst[len - 1] == L'0',
	      "Failed to properly seed dst; first part of long char dst[%d] = `%c'.",
	      len - 1, (char)dst[len - 1]);

  /*
   * In this case, we supposedly didn't properly terminate dst with a
   * NULL byte. Thus, wcslcat() is supposed to leave this last byte
   * untouched.
   */
  ret = wcslcat(dst, src, len);
  fail_unless(ret == len + wcslen(src),
	      "Expected ret == len + wcslen(src), but ret = %d and len + wcslen(src) = %d",
	      ret, len + wcslen(src));
  fail_unless(!wcsncmp(dst, L"12345678901234567890", len),
	      "Expected dst not to be modified, it was");
  fail_unless(dst[len - 1] != '\0',
	      "Expected last byte of dst not to be '\0', but it was '\0'");
}
END_TEST

START_TEST (wcslcat_zero)
{
  wchar_t *dst;
  const wchar_t *src = L"pizza";
  size_t len;
  size_t ret;

  len = 0;

  dst = malloc(sizeof(wchar_t) * len);
  /*
   * It is valid for dst to be either NULL or some valid pointer at
   * this point.
   */

  ret = wcslcat(dst, src, len);
  fail_unless(ret == wcslen(src),
	      "Expected ret to be %d, was instead %d.",
	      0, ret);

  free(dst);

  len ++;
  dst = malloc(sizeof(wchar_t) * len);
  fail_unless(dst != 0,
	      "Unable to malloc() %d bytes", len);

  dst[0] = L'\0';

  ret = wcslcat(dst, src, len);
  fail_unless(ret == wcslen(src),
	      "Expected ret to be %d, was instead %d.",
	      wcslen(src), ret);
  fail_unless(dst[0] == L'\0',
	      "Destination string was not properly null-terminated, last wide char's first byte is `%c'.",
	      dst[0]);
  fail_unless(wcslen(dst) == len - 1,
	      "Destination string's length is %d, expected length of %d",
	      wcslen(dst), len - 1);

  free(dst);
}
END_TEST

Suite *wcslcat_suite()
{
  Suite *s;
  TCase *tc;

  s = suite_create("wcslcat");

  tc = tcase_create("empty");
  tcase_add_test(tc, wcslcat_empty);
  suite_add_tcase(s, tc);

  tc = tcase_create("append");
  tcase_add_test(tc, wcslcat_append);
  suite_add_tcase(s, tc);

  tc = tcase_create("fill");
  tcase_add_test(tc, wcslcat_fill);
  suite_add_tcase(s, tc);

  tc = tcase_create("clip");
  tcase_add_test(tc, wcslcat_clip);
  suite_add_tcase(s, tc);

  tc = tcase_create("overfull");
  tcase_add_test(tc, wcslcat_overfull);
  suite_add_tcase(s, tc);

  tc = tcase_create("zero");
  tcase_add_test(tc, wcslcat_zero);
  suite_add_tcase(s, tc);

  return s;
}
